from expr_parser.parser import Parser
from expr_parser.operators import default


__all__ = [
    "Parser",
    "parse",
    "evaluate",
    "function"
]


_default_parser = Parser.default()


def parse(expr: str):
    return _default_parser.parse(expr)


def evaluate(expr: str):
    return parse(expr).eval()


def function(expr: str):
    tree = parse(expr)
    return lambda x: tree.eval(x=x)



import re as _re
import math as _math
from typing import Dict as _Dict

from expr_parser.operators.base import Operator as _Operator
from expr_parser.tree import (Tree as _Tree,
                              Variable as _Var,
                              Constant as _Const,
                              tree_from_list as _tree_from_list)
from expr_parser.operators.default import (Implicit as _Implicit,
                                           Add as _Add,
                                           Sub as _Sub,
                                           Mul as _Mul,
                                           Div as _Div,
                                           Pow as _Pow)


class Parser:

    @staticmethod
    def default():
        parser = Parser()
        parser.add_bracket("(", ")")
        for op in (_Implicit, _Add, _Sub, _Mul, _Div, _Pow):
            parser.add_operator(op)
        parser.constants = {
            "pi": _math.pi,
            "e": _math.e,
            "i": 1j,

            "sin": _math.sin,
            "cos": _math.cos,
            "tan": _math.tan,
            "log": _math.log
        }
        return parser

    def __init__(self):
        self.brackets: _Dict[str, str] = {}
        self.operators: _Dict[str, _Operator] = {}
        self.constants: _Dict = {}

    def add_operator(self, operator: _Operator):
        self.operators[operator.symbol] = operator

    def add_bracket(self, opening, closing):
        self.brackets[opening] = closing

    def _is_opening_bracket(self, string: str) -> bool:
        return string in self.brackets.keys()

    def _is_closing_bracket(self, string: str) -> bool:
        return string in self.brackets.values()

    def _is_bracket(self, string: str) -> bool:
        return self._is_opening_bracket(string) or self._is_closing_bracket(string)

    def _is_operator(self, string: str) -> bool:
        return string in self.operators.keys()

    def _is_valid_token(self, s: str):
        return _is_numeric(s) or self._is_bracket(s) or self._is_operator(s) or s.isidentifier()

    def _create_token(self, s: str):
        if self._is_bracket(s):
            return s
        elif self._is_operator(s):
            return self.operators[s]
        elif s.isidentifier():
            return _Var(s)
        elif _is_numeric(s):
            return _parse_numeric_const(s)
        else:
            raise ValueError(f"Unknown token: '{s}'")

    def _tokenize(self, string: str):
        token = ""
        for char in string:
            if char == " ":
                if token != "":
                    yield self._create_token(token)
                    token = ""
            elif self._is_valid_token(token + char):
                token += char
            else:
                yield self._create_token(token)
                token = char

        if token != "":
            yield self._create_token(token)

    def _tokenize_with_implicit(self, string):
        iterator = self._tokenize(string)

        try:
            token = next(iterator)
        except StopIteration:
            return

        for next_token in iterator:
            yield token
            if isinstance(token, (_Const, _Var)) and (
                    isinstance(next_token, (_Const, _Var)) or self._is_opening_bracket(next_token)):
                yield self.operators["==IMPLICIT=="]
            token = next_token

        yield token

    def _group(self, tokens, process_scope=lambda x: x):
        """
        Group a stream of tokens into groups/scopes by brackets
        """
        result = []
        stack = []
        scope = result

        for token in tokens:
            # Got a opening bracket
            if self._is_opening_bracket(token):

                # Push current scope and opening bracket on stack
                stack.append((token, scope))

                # Create and enter new scope
                scope.append([])
                scope = scope[-1]

            # Got a closing bracket
            elif self._is_closing_bracket(token):
                # Check if there are any open brackets to close
                if len(stack) == 0:
                    raise SyntaxError("Missing opening bracket")

                # Enter the outer scope and get the opening bracket being closed
                opening, scope = stack.pop()

                # Check if the opening bracket matches the closing one
                if self.brackets[opening] != token:
                    raise SyntaxError("Got mismatching brackets")

                # Process just finished scope
                scope[-1] = process_scope(scope[-1])

            # Got no bracket
            else:
                scope.append(token)

        # Check if all brackets have been closed
        if len(stack) > 0:
            raise SyntaxError("Missing closing bracket")

        # Process and returned finished outermost scope
        return process_scope(result)

    def parse(self, expr: str):
        if self._is_operator("==IMPLICIT=="):
            root = self._group(self._tokenize_with_implicit(expr), process_scope=_tree_from_list)
        else:
            root = self._group(self._tokenize(expr), process_scope=_tree_from_list)
        return _Tree(root, self.constants)


__numeric_re = _re.compile(r"^(\d+)?(\.\d*)?$")


def _is_numeric(string):
    return __numeric_re.match(string) is not None and string != "."


def _parse_numeric_const(string) -> _Const:
    try:
        return _Const(int(string))
    except ValueError:
        pass
    try:
        return _Const(float(string))
    except ValueError:
        pass
    raise ValueError(f"'{string}' is neither an int or float")



"""
Provide implementation for the basic math operators:

+, -, *, /, ^
"""

import operator as _op

from expr_parser.operators.base import Operator as _Operator


def _implicit(x, y):
    if callable(x) and callable(y):
        return lambda i: x(y(i))
    elif callable(x):
        return x(y)
    elif callable(y):
        return lambda i: x * y(i)
    else:
        return x * y


Implicit = _Operator(
    symbol="==IMPLICIT==",
    priority=10,
    binary=_implicit
)

Add = _Operator(
    symbol="+",
    priority=0,
    unary=lambda x: x,
    binary=_Operator.handle_callables(_op.add)
)

Sub = _Operator(
    symbol="-",
    priority=0,
    unary=_op.neg,
    binary=_Operator.handle_callables(_op.sub)
)

Mul = _Operator(
    symbol="*",
    priority=10,
    binary=_Operator.handle_callables(_op.mul)
)

Div = _Operator(
    symbol="/",
    priority=10,
    binary=_Operator.handle_callables(_op.truediv)
)

Pow = _Operator(
    symbol="^",
    priority=20,
    binary=_Operator.handle_callables(_op.pow)
)



from typing import (TypeVar as _TypeVar,
                    Dict as _Dict,
                    Set as _Set,
                    Union as _Union)

from expr_parser.operators.base import Operator as _Operator
from expr_parser.error import MissingVariable


_Numeric = _TypeVar("_Numeric", int, float, complex)


class Tree:

    __slots__ = ["root", "initial_namespace"]

    def __init__(self, root: "_Node", initial_namespace=None):
        self.root = root
        if initial_namespace is None:
            self.initial_namespace = {}
        else:
            self.initial_namespace = initial_namespace

    def _get_effective_namespace(self, namespace: _Dict):
        return dict(self.initial_namespace, **namespace)

    @property
    def used_variables(self):
        return self.root.used_variables

    def eval(self, **namespace: _Numeric) -> _Numeric:
        namespace = self._get_effective_namespace(namespace)
        for var in self.used_variables:
            if var not in namespace:
                raise MissingVariable(var)
        return self.root._eval(namespace)

    def simplified(self, **namespace: _Numeric) -> "Tree":
        namespace = self._get_effective_namespace(namespace)
        root = self.root._eval_partial(namespace)
        if isinstance(root, _Node):
            return Tree(root)
        else:
            return Tree(Constant(root))

    def __call__(self) -> _Numeric:
        return self.eval()


class _Node:

    def _eval(self, namespace: _Dict[str, _Numeric]) -> _Numeric:
        raise NotImplementedError()

    def _eval_partial(self, namespace: _Dict[str, _Numeric]) -> _Union["_Node", _Numeric]:
        raise NotImplementedError()

    @property
    def used_variables(self) -> _Set[str]:
        raise NotImplementedError()

    def _is_none(self) -> bool:
        """
        Overwritten in Constant to check if it's internal value is None.

        Operation uses this function to determine if it's a unary or binary operation.
        """
        return False


class Operation(_Node):

    __slots__ = ["operator", "x", "y"]

    def __init__(self, operator: _Operator, x: _Node, y: _Node = None):
        self.operator = operator
        self.x = x
        if y is None:
            self.y = Constant(None)
        else:
            self.y = y
        self._used_variables = self.x.used_variables | self.y.used_variables

    def _eval(self, namespace: _Dict[str, _Numeric]) -> _Numeric:
        x = self.x._eval(namespace)
        y = self.y._eval(namespace)

        return self.operator(x, y)

    def _eval_partial(self, namespace: _Dict[str, _Numeric]) -> _Union[_Node, _Numeric]:
        x = self.x._eval_partial(namespace)
        y = self.y._eval_partial(namespace)

        if not isinstance(x, _Node) and not isinstance(y, _Node):
            return self.operator(x, y)
        if not isinstance(x, _Node):
            x = Constant(x)
        if not isinstance(y, _Node):
            y = Constant(y)
        return Operation(self.operator, x, y)

    @property
    def used_variables(self) -> _Set[str]:
        return self._used_variables

    def __repr__(self):
        return f"Operation({self.operator}, {repr(self.x)}, {repr(self.y)})"

    def __str__(self):
        if self.y._is_none():
            return f"({self.operator} {self.x})"
        else:
            return f"({self.x} {self.operator} {self.y})"


class Constant(_Node):

    __slots__ = ["value"]

    def __init__(self, value: _Numeric):
        self.value = value

    def _is_none(self):
        return self.value is None

    def _eval(self, namespace: _Dict[str, _Numeric]) -> _Numeric:
        return self.value

    def _eval_partial(self, namespace: _Dict[str, _Numeric]) -> _Union[_Node, _Numeric]:
        return self.value

    @property
    def used_variables(self) -> _Set[str]:
        return set()

    def __repr__(self):
        return f"Const({repr(self.value)})"

    def __str__(self):
        return str(self.value)


class Variable(_Node):

    __slots__ = ["name"]

    def __init__(self, name: str):
        self.name = name

    def _eval(self, namespace: _Dict[str, _Numeric]) -> _Numeric:
        if self.name in namespace:
            return namespace[self.name]
        else:
            raise MissingVariable(self.name)

    def _eval_partial(self, namespace: _Dict[str, _Numeric]) -> _Union[_Node, _Numeric]:
        try:
            return self._eval(namespace)
        except MissingVariable:
            return Variable(self.name)

    @property
    def used_variables(self) -> _Set[str]:
        return {self.name}

    def __repr__(self):
        return f"Var({repr(self.name)})"

    def __str__(self):
        return str(self.name)


def tree_from_list(lst):
    def consume_operator(i):
        op = lst[i]
        if i+1 == len(lst) or isinstance(lst[i+1], _Operator):
            raise SyntaxError(f"Operator '{op}' is missing its operand")
        right = lst[i+1]

        if i > 0 and not isinstance(lst[i-1], _Operator):
            left = lst[i-1]
        else:
            left = None

        if left is None:
            return lst[:i] + [Operation(op, right)] + lst[i+2:]
        else:
            return lst[:i-1] + [Operation(op, left, right)] + lst[i + 2:]

    while len(lst) > 1:
        if isinstance(lst[0], _Operator):
            index = 0
        else:
            index = 1

        while index + 2 < len(lst) \
                and isinstance(lst[index+2], _Operator) \
                and lst[index].priority < lst[index+2].priority:
            index += 2

        lst = consume_operator(index)

    return lst[0]



"""
Implement a dice operator:

2d6 rolls two 6-sided dice and adds their results
"""

from random import randint as _randint

from expr_parser.operators.base import Operator as _Operator


Dice = _Operator(
    symbol="d",
    priority=100,
    unary=lambda x: _randint(1, x),
    binary=lambda x, y: sum((_randint(1, y) for i in range(x)))
)



import sys

from expr_parser import evaluate, parse


def main(args):
    if len(args) < 2:
        print(f"Usage: {args[0]} <EXPRESSION> [<KEY>=<VALUE> ...]")
        print(f"")
        print(f"This program evaluates a arithmetic expression.")
        print(f"Any number of key-value-pairs can be given to use for variables:")
        print(f"    {args[0]} x^2+y x=2 y=1")
        print(f"    5")
        exit(1)
    elif len(args) == 2:
        print(evaluate(args[1]))
    else:
        namespace = dict((
            (key, evaluate(value))
            for key, value in map(
                lambda x: x.split("="),
                args[2:]
            )
        ))
        print(parse(args[1]).eval(**namespace))
    exit(0)


if __name__ == "__main__":
    main(sys.argv)



from typing import Callable as _Callable
from functools import wraps as _wraps


class Operator:

    def __init__(self, symbol: str, priority: int, unary: _Callable = None, binary: _Callable = None):
        self.symbol = symbol
        self.priority = priority
        self.unary = unary
        self.binary = binary

    def __call__(self, x, y=None):
        if y is None:
            return self.unary(x)
        else:
            return self.binary(x, y)

    @staticmethod
    def handle_callables(func):
        @_wraps(func)
        def new_func(x, y):
            if callable(x) and callable(y):
                return lambda i: func(x(i), y(i))
            elif callable(x):
                return lambda i: func(x(i), y)
            elif callable(y):
                return lambda i: func(x, y(i))
            else:
                return func(x, y)
        return new_func

    def __repr__(self):
        return f"Operator({self.symbol})"

    def __str__(self):
        return self.symbol
